#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "define.h"

#define SOCKET_ERROR   ((int)-1)
#define X_init 3
#define Y_init 10
#define LARGHEZZA 70
#define ALTEZZA_1 10
#define ALTEZZA_2 12
#define BORDO_TABELLA BG_BLUE
#define COL_SCRITTE YELLOW
#define BOLD	1
#define BLINK	5
#define REVERSE	7
#define NORMAL	0
#define cprintf printf


//======================================================== Funzioni di GraFica ================

/* Enumerazione dei colori */
enum fgcolor {BLACK = 30 , RED , GREEN , BROWN , BLUE , MAGENTA , CYAN , WHITE , YELLOW};
enum bgcolor {BG_BLACK = 40 , BG_RED , BG_GREEN , BG_BROWN , BG_BLUE , BG_MAGENTA , BG_CYAN , BG_WHITE};

//---------------------------------------------------------------------------------------------
void gotoxy(int x , int y);
        /* Permette di spostarsi in un certo punto dello schermo definito dalle coordinate
                x ed y
                - x = ascisse
                - y = ordinata */
//---------------------------------------------------------------------------------------------
void clrscr();
        /* Pulisce lo schermo in modo da rendere pi gradevole la grafica */
//---------------------------------------------------------------------------------------------
void _textattr(int attrib);
        /* Imposta il testo, sia foreground che background, con un certo
                attributo attrib */
//---------------------------------------------------------------------------------------------
int grafica_iniziale(char *richiesto , char *salvato , int filesize);
        /* Scriviamo su stdout, appena pulito, il titolo, la tabella e altre scritte
                - richiesto = file chiesto dal server al proxy
                - salvato = nome con cui il client salva il file scaricato dal proxy
                - filesize = dimensione del file */
//---------------------------------------------------------------------------------------------
void disegna_asterischi(int filesize , int segmento);
        /* disegna dei caratteri in una tabella e rappresenta l'avanzamento del download del file
                - filesize = dimensione del file
                - segmento = n del segmento corrente*/
//---------------------------------------------------------------------------------------------
void percentuale (int filesize , int byte);
        /* stampa la percentuale del file ricevuto durante la sua trasmissione
                - filesize = dimensione del file
                - byte = n del byte corrente */
//---------------------------------------------------------------------------------------------
int velocita (struct timeval *inizio , struct timeval *fine , int segmento);
        /* scrive su stdout la velocit media con cui riesce a scaricare il file; viene aggiornato ogni volta che viene chiamata,
                - inizio = struttura timeval che contiene il tempo iniziale
                - fine = struttura timeval che contiene il tempo finale
                - segmento = n del segmento corrente */
//---------------------------------------------------------------------------------------------
void rimanenza (int segmento , int filesize , float velocita);
        /* Scrive su stdout quanti secondi *dovrebbero* mancare al termine del trasferimento del file
                - segmento = n del segmento corrente
                - filesize = dimensione del file
                - velocita = valore di ritorno della funzione velocita(), indica la vel. di trasferimento del file */
//=============================================================================================


//============================================================ Lettura e Scrittura ============
ssize_t writen (int fd, const void *buf, size_t n){
        size_t nleft;
        ssize_t nwritten;
        char *ptr;

        ptr = (char*)buf;
        nleft = n;
        while (nleft > 0) {
                if ( (nwritten = write(fd , ptr , nleft)) <= 0) {
                        if (errno == EINTR)
                                nwritten = 0;
                        else
                                return -1;
                        }
                nleft -= nwritten;
                ptr += nwritten;
                }
        return n;
        }

//----------------------------------------------------------------------------------------------------------

int readn(int fd, char *ptr, int nbytes, int mode);
/* Usata per lettura "sicura"; finch non riceviamo nbytes cicla, altrimenti restituisce un errore;
        - fd = file descriptor da cui ricevere i dati
        - ptr = puntatore alla stringa dove andremo a salvare i dati entrati
        - nbytes = n di byte da leggere
        - mode = se 1 stampa i dati relativi alla ricezione del file sullo schermo (velocit, percentuale, etc)
                altrimenti non stampa nulla */

int readn(int fd, char *ptr, int nbytes, int mode){
        int nleft , nread , ris;
        struct timeval end ,inizio;

        nleft = nbytes;
        gettimeofday(&inizio , NULL);
        while(nleft > 0){
                nread = read(fd , ptr , 1);
                if (mode){                                                 /* parte grafica */
                        disegna_asterischi(nbytes , nbytes-nleft);         /* barra di incremento */
                        percentuale(nbytes , nbytes-nleft + 1);            /* percentuale ricevuta */
                        gettimeofday(&end , NULL);                         /* per il calcolo del tempo impiegato */
                        ris = velocita(&inizio , &end , nbytes-nleft);     /* velocit di trasmissione dati */
                        rimanenza(nbytes , nbytes-nleft , ris);            /* tempo stimato rimanente */
                        }
                if(nread<0)
                        return(-1);
                else
                        if(nread == 0) {
                                return(0);
                                break;
                                }
                nleft -= nread;
                ptr   += nread;
                }
        return(nbytes);
        }
//===========================================================================================




//-------------------------------------------------------------------------------------------
void usage(char *name);
/* stampa un messaggio di errore su stdout in caso di errato numero di parametri passati al programma */

void usage(char *name){
        printf ("Error: invalid number of args \n");
        printf ("Usage: %s server_address server_port file output_file\n" , &name);
        fflush (stdout);
        }
//-------------------------------------------------------------------------------------------
GetRequestHeader makerequest (char * filename);
/* Crea un pacchetto di tipo GetRangeRequest da spedire al server per richiedere un file
        - filename = nome del file richiesto */

GetRequestHeader makerequest (char * filename){
        GetRequestHeader request;
        memset(&request , 0 , sizeof(request));
        request.request_code = 'G';                  /* tipo del pacchetto: G [GET] */
        if (strlen(filename) > FILENAMELEN - 1){
                _textattr(RED);                      /* controllo dell'errore */
                _textattr(BOLD);
                printf("Errore: Nome file troppo grande\n");
                _textattr(NORMAL);
                exit(1);
                }
        else
                strcpy(request.filename , filename);  /* se tutto ok copio il nome del file nel pacchetto */
        return request;
        }
//-----------------------------------------------------------------------------------------
void segnale();
/* Handler per gestire l'interrupt generato dall'utente che ha premuto CTRL + C;
        sposta il cursore in basso e poi termina l'esecuzione del programma, per
        non lasciare delle scritte in mezzo allo schermo */

void segnale(){
        gotoxy(1,29);
        signal(SIGINT , SIG_DFL);
        signal(SIGTERM , SIG_DFL);
        raise(SIGINT);
        }
//------------------------------------------------------------------------------------------
void save(char nomefile[],char contenutofile[],int size);
/*Salva i dati che gli arrivano dal server in un file
        - nomefile - nome del file in cui salvare i dati in arrivo dal server
        - contenutofile - dati da salvare nel file
        - size - dimensione del file */

void save(char nomefile[],char contenutofile[],int size){
        int n;
        FILE *fp;

        fp = fopen(nomefile , "w");                     /* apriamo un file in scrittura */
        n = fwrite(contenutofile , 1 , size , fp);      /* e ci scriviamo i nostri dati */
        if (n < 0)
                printf("Errore nella scrittura del file\n");
        fclose(fp);
        }
//-------------------------------------------------------------------------------------------



//==============================================================================================
//                                                                       << M A I N >>
//==============================================================================================
int main(int argc, char *argv[]){
        struct sockaddr_in Serv;
        struct timeval inizio;
        int sockfd , OptVal , msglen , Fromlen , ris;
        int n , nread , nwrite , len , lung;               /* definizioni variabili */
        char file[MAXFILESIZE] , tipo;
        ErrorReplyHeader err;
        GetReplyHeader grep;
        GetRequestHeader greq;
        int *ris3;

        signal(SIGINT , segnale);
        _textattr(NORMAL);
        clrscr();
        if(argc!=5){                            /* controllo sull'immissione dati */
                usage(argv[0]);
                exit(1);
                }
        sockfd = socket(AF_INET, SOCK_STREAM, 0);     /* inizializziamo un nuovo socket */
        if (sockfd == SOCKET_ERROR) {                 /* e controlliamo se tutto  andato a buon fine */
                gotoxy(0,22);
                _textattr(RED);
                printf ("socket() failed, Err: %d \"%s\"\n" , errno , strerror(errno));
                _textattr(WHITE);
                exit(1);
                }

        OptVal = 1;     /* settiamo le opzioni del socket <riutilizzo dell'indirizzo> */
        ris = setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&OptVal, sizeof(OptVal));
        if (ris == SOCKET_ERROR)  {
                gotoxy(0,22);                 /* controllo dell'errore */
                _textattr(RED);
                printf ("setsockopt() SO_REUSEADDR fallita, Errore: %d \"%s\"\n", errno,strerror(errno));
                _textattr(WHITE);
                exit(1);
                }

        memset( &Serv , 0 , sizeof(Serv) );
        Serv.sin_family = AF_INET;                              /* immissione dei parametri di connessione al server */
        if (inet_addr(argv[1]) == inet_addr("255.255.255.255"))
                return - 1;
        Serv.sin_addr.s_addr = inet_addr(argv[1]);
        Serv.sin_port = htons(atoi(argv[2]));

        ris = connect(sockfd, (struct sockaddr*) &Serv, sizeof(Serv));   /* tentiamo di connetterci al server */
        if (ris == SOCKET_ERROR)  {
                _textattr(RED);                                          /* controllo dell'errore */
                printf ("connect() fallita, Errore: %d \"%s\"\n",errno,strerror(errno));
                _textattr(WHITE);
                exit(1);
                }

        greq = makerequest(argv[3]);                    /* incapsulamento della richiesta GET */
        ris = writen(sockfd , &greq , sizeof(greq));    /* scrittura della richiesta sul server */
        if (ris == -1)                                  /* controllo dell'errore */
                return ris;

        ris = readn(sockfd , (char*) &grep , sizeof(grep.respons_code) , 0);  /* lettura della risposta alla GET dal server */
        if (ris == -1)
                return ris;

        tipo = grep.respons_code;       /* controlliamo il tipo del pacchetto appena ricevuto */
        if (tipo == 'G'){               /* ricevuto GET, quindi lettura della lunghezza del file */
                ris = readn(sockfd ,((char*) &grep) + sizeof(grep.respons_code) , (sizeof(GetReplyHeader) - sizeof(grep.respons_code)) , 0);
                if (ris == -1){
                        _textattr(RED);
                        printf("Errore in ricezione\n");      /* controllo dell'errore */
                        _textattr(NORMAL);
                        return -1;
                        }
                }
        else
                if(tipo == 'E') {        /* ricevuto ERRORE, quindi lettura del messaggio d'errore */
                        ris = readn(sockfd , ((char*) &err) + sizeof(err.error_code) , (sizeof(ErrorReplyHeader) - sizeof(err.error_code)) , 0);
                        _textattr(RED);
                        printf("Ricevuto errore: %s\n" ,err.msg);
                        _textattr(NORMAL);
                        return -1;
                        }
                else
                        return -1;

        lung = atoi(grep.bodysize);                     /* lunghezza del file */
        grafica_iniziale(argv[3] , argv[4] , lung);     /* disegniamo la tabella e i titoli */
        gettimeofday(&inizio , NULL);                   /* salviamo l'ora iniziale per il controllo del tempo nella parte grafica */
        ris = readn(sockfd , (char*) &file , lung - 1 , 1);  /* leggiamo il file di lunghezza <lung> */
        if (ris == -1)
                return ris;
        save(argv[4] , file , lung - 1);                 /* salviamo i dati ricevuti in un file */
        close(sockfd);                                   /* chiudiamo il socket verso il server */

        _textattr(NORMAL);
        gotoxy(1 , 29);                 /* setto il font alla normalit, in modo da non influenzare */
        printf("\n");                   /* altre applicazioni */

        return 0;
        }
//================================================================================================== FINE MAIN =========
















//======================================================= FUNZIONI X LA GRAFICA =================================================


void gotoxy(int x , int y){
        char es[50];
        char posx[25];
        char posy[25];

        sprintf(posx , "%d" , x);
        sprintf(posy , "%d" , y);
        es[0] = '\0';
        strcat(es , "\x1B[");
        strcat(es , posy);
        strcat(es , "d");
        strcat(es , "\x1B[");
        strcat(es , posx);
        strcat(es , "G");
        printf("%s" , es);
        }

void clrscr(void){
        printf("\x1B[2J"
                "\x1B[0d"
                "\x1B[0G");
        }


void _textattr(int attrib){
        if (attrib == YELLOW){
                _textattr(BROWN);
                _textattr(BOLD);
                }
        else{
                char es[50];
                char attr[25];
                sprintf (attr , "%d" , attrib);
                es[0] = '\0';
                strcat (es , "\x1B[");
                strcat (es , attr);
                strcat (es , "m");
                printf ("%s" , es);
                }
        }


int grafica_iniziale(char *richiesto, char *salvato, int filesize){
        int x,y;

        clrscr();
        _textattr(35);
        _textattr(BOLD);
        gotoxy(15,2);
        printf( "Progetto di Laboratorio"
                " di Programmazione di Reti\n");
        gotoxy(15,1);                                   /* titolo  del progetto*/
        printf("----------------------"
                "---------------------------\n");
        gotoxy(15,3);
        printf("----------------------"
                "---------------------------\n");

        _textattr(CYAN);
        gotoxy(30,4);
        printf("Prof. Fabio Panzieri\n");               /* All'attenzione di: */
        gotoxy(30,5);
        printf("Prof. Vittorio Ghini\n");

        _textattr(GREEN);
        _textattr(BOLD);                                /* file richiesto: <nome_file_dst> */
        gotoxy(13,7);                                   /*   salva come: <nome_file_rcv>   */
        printf("file richiesto: %s     -     "
                "salva come: %s",richiesto,salvato);

        gotoxy(25,8);
        _textattr(NORMAL);                              /* dimensione file : <n> byte */
        printf("dimensione file: %d byte", filesize);

        _textattr(BORDO_TABELLA);                       /* disegno i bordi della tabella */
        for (x = X_init , y = Y_init ; x < X_init + LARGHEZZA + 4 ; x++) {
                gotoxy(x , y);                             /*  ^^^ 4 = 2 caratteri di inizio + 2 di fine tabella */
                printf(" ");
                gotoxy(x , y + ALTEZZA_2 - ALTEZZA_1);
                printf(" ");
                }
        for (x = X_init , y = Y_init + 1 ; y < Y_init + ALTEZZA_2 - ALTEZZA_1 + 1 ; y++) {
                gotoxy(x , y);
                printf("  ");
                gotoxy(x + LARGHEZZA + 2 , y);
                printf("  ");
                }

        _textattr(COL_SCRITTE);            /* inserisco le scritte relative alla */
        _textattr(BG_BLACK);          /*  trasmissione dei dati             */
        gotoxy(15,14);
        printf("Percentuale:           0.00    %c",'%');   /* percentuale di file scaricato */
        gotoxy(15,15);
        printf("Velocit media:        0.00    B/s");      /* velocit in ricezione */
        gotoxy(15,16);
        printf("Tempo rimanente:       0       s");        /* secondi rimanenti per terminare il download */

        gotoxy(77,25);
        _textattr(BG_GREEN);         /* FORZA ITALIA! :) */
        printf(" ");
        _textattr(BG_WHITE);
        printf(" ");
        _textattr(BG_RED);
        printf(" ");
        gotoxy(0,25);
        _textattr(BG_GREEN);
        printf(" ");
        _textattr(BG_WHITE);
        printf(" ");
        _textattr(BG_RED);
        printf(" ");

        _textattr(YELLOW);
        _textattr(BG_BLACK);
        gotoxy(15,25);
        printf("A. Michetti");       /* Autori... */
        gotoxy(35,25);
        printf("R. Valenti");
        gotoxy(55,25);
        printf("A. Pegoraro");

        _textattr(NORMAL);          /* reimpostiamo il font alla normalit */
        printf("\n\n");
}


void disegna_asterischi (int filesize , int segmento){
        int riga;
        int colonna;

        for (riga = ALTEZZA_1 + 1 ; riga < ALTEZZA_2 ; riga++)
                for (colonna = (segmento * LARGHEZZA / filesize) - 1; colonna < ((segmento + 1) * LARGHEZZA / filesize) - 1 ; colonna++) {
                        gotoxy (colonna + X_init + 3 , riga);      /* controllo per riempire tutta la tabella */
                        _textattr (COL_SCRITTE);
                        printf("%c" , 130);   /* il carattere 130 in ASCII corrisponde al quadratino */
                        }
        _textattr (WHITE);
        }

        void percentuale (int filesize , int byte){
        float perc;                    /*                        byte  *  100             */
        perc = byte * 100;             /*       percentuale = ------------------          */
        perc /= filesize;              /*                          filesize               */

        _textattr (COL_SCRITTE);
        gotoxy (32,14);                /* Andiamo a scrivere la percentuale sullo schermo */
        printf ("%10.2f",perc);

        if (perc >= 100)
                {
                gotoxy (50,14);
                _textattr (BLINK);
                printf ("Download completato!!!\a");    /* tutto  andato a buon fine, almeno si spera! :) */
                _textattr (NORMAL);
                }
        _textattr (NORMAL);
        }

int velocita (struct timeval *inizio , struct timeval *fine , int segmento){
        float tempo , vel;

        vel = segmento;
        tempo = ((fine)->tv_sec) - ((inizio)->tv_sec) + 1;
        if (tempo < 0)                 /*               byte scritti   */
                tempo *= -1;           /* velocit = ----------------- */
        vel /= tempo;                  /*             tempo impiegato  */

        gotoxy(32 , 15);
        _textattr(COL_SCRITTE);              /* scrittura della velocit attuale su schermo */
        if (vel < 1000L)
                printf("%10.2f  B/s  " , vel);                      /* in  byte / sec */
        else
                printf("%10.2f kB/s  " , vel / 1000L);              /* in Kbyte / sec */
        _textattr(NORMAL);
        gotoxy(1,29);     /* posizioniamo il puntatore in fondo */
        return vel;
        }

void rimanenza (int segmento , int filesize , float velocita){
        float secondi , byte_scritti , finale;

        if (segmento != 0)
                byte_scritti = segmento;
        else
                byte_scritti = segmento + 1;         /*                     filesize * t_impiegato                 */
        secondi = byte_scritti / velocita;           /*  tempo_rimanente = ------------------------  - t_impiegato */
        if (secondi < 0)                             /*                           byte_scritti                     */
                secondi *= -1;
        finale = (secondi * filesize / byte_scritti) - secondi;

        gotoxy(32 , 16);
        _textattr(COL_SCRITTE);     /* scrittura del tempo mancante per terminare la trasmissione */
        printf("%10.0f" , - finale);      /* del file */
        _textattr(NORMAL);

        gotoxy(1,29);
        }
